#!/usr/bin/env python
#   -*- encoding: UTF-8 -*-
#
#   :Author: tundish
#   :Date: 2007-02-15
#   :Copyright: 2007 Thuswise Ltd
#   :Description:   test-sprinkle runs unit tests on sprinkle
#                   coded for Python 2.3.4 and above
#   :Contact: tundish at thuswise . org . uk

#   This file is part of the sprinkle distribution

#   Sprinkle is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>

import os
import sys
import time
import signal
import socket
import unittest
import optparse
import tempfile
from ConfigParser import SafeConfigParser, DuplicateSectionError
 
import sprinkle

USER_ROOT = "u-test"
QUEUE_ROOT = "q-test"

# Helper functions

def makeDir(root):
    """ A quick way to make a new directory tree """
    try:
        os.makedirs(root)
    except OSError:
        pass

def nukeDir(root):
    """ Remove an entire directory tree """
    for root, dirs, files in os.walk(root, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
    try:
        os.rmdir(root)
    except OSError:
        pass

def makeUser(user, passwd, groups, ack="auto"):
    """ A white-box way of making a new user """
    cp = SafeConfigParser()
    um = sprinkle.UserManager(USER_ROOT)
    cp.set("DEFAULT", "passcode", passwd)
    for i in groups:
        try:
            cp.add_section(i)
        except DuplicateSectionError:
            pass
        cp.set(i, "ack", ack)
    ur = um._userRef(user)
    makeDir(os.path.dirname(ur))
    f = file(ur,'w')
    cp.write(f)
    f.close()

class TestFormat(unittest.TestCase):
    """ Concerning the creation of messages"""

    def test_001(self):
        f = sprinkle.connect(login="siegfried",passcode="canine")
        self.failUnlessEqual(str(f),
        ("CONNECT\nlogin: siegfried\npasscode: canine\n\n\x00"))
 
    def test_002(self):
        f = sprinkle.connected(session="1066")
        self.failUnlessEqual(str(f),
        ("CONNECTED\nsession: 1066\n\n\x00"))
 
    def test_003(self):
        f = sprinkle.subscribe(
        destination="gardening",ack="client")
        self.failUnlessEqual(str(f),
        ("SUBSCRIBE\nack: client\ndestination: gardening\n\n\x00"))
        f.requestReceipt("m003")        
        self.failUnlessEqual(str(f),
        ("SUBSCRIBE\nack: client\ndestination: gardening\n"
        "receipt: m003\n\n\x00"))
 
    def test_004(self):
        f = sprinkle.receipt(id="m003")
        self.failUnlessEqual(str(f),
        ("RECEIPT\nreceipt-id: m003\n\n\x00"))
 
    def test_005(self):
        f = sprinkle.begin(transaction="t4")
        self.failUnlessEqual(str(f),
        ("BEGIN\ntransaction: t4\n\n\x00"))
 
    def test_006(self):
        f = sprinkle.commit(transaction="t4")
        self.failUnlessEqual(str(f),
        ("COMMIT\ntransaction: t4\n\n\x00"))
 
    def test_007(self):
        f = sprinkle.abort(transaction="t4")
        self.failUnlessEqual(str(f),
        ("ABORT\ntransaction: t4\n\n\x00"))
 
    def test_008(self):
        msg = "This is a message"
        f = sprinkle.message(msg, destination="gardening", id="m003")
        self.failUnlessEqual(str(f),
        ("MESSAGE\ncontent-length: %d\n" 
        "destination: gardening\nmessage-id: m003\n\n%s\x00" %
        (len(msg),msg)))
 
    def test_009(self):
        f = sprinkle.ack(id="m003")
        self.failUnlessEqual(str(f),
        ("ACK\nmessage-id: m003\n\n\x00"))
 
    def test_010(self):
        msg = "This is a message"
        f = sprinkle.send(msg, destination="gardening")
        self.failUnlessEqual(str(f),
        ("SEND\ncontent-length: %d\n" 
        "destination: gardening\n\n%s\x00" %
        (len(msg),msg)))
 
    def test_011(self):
        msg = "I was just cleaning it and it went off in my hand"
        f = sprinkle.error(msg, message="Ooops!")
        self.failUnlessEqual(str(f),
        "ERROR\ncontent-length: %d\nmessage: Ooops!\n\n%s\x00" %
        (len(msg),msg))
 
    def test_012(self):
        """ NB: Minor incompatibility from spec (extra newline) """
        f = sprinkle.disconnect()
        self.failUnlessEqual(str(f),"DISCONNECT\n\n\n\x00")
 
class TestDeframer(unittest.TestCase):
    """ Concerning the Deframer class"""

    def setUp(self):
        self.d = sprinkle.Deframer()

    def test_001(self):
        """ Deframe a single frame"""
        f = sprinkle.connect(login="siegfried",passcode="canine")
        self.d.load(str(f))
        self.failUnlessEqual(self.d.next(),
        ("CONNECT",{"login": "siegfried","passcode": "canine"},""))

    def test_002(self):
        """ Deframe multiple frames """
        a = sprinkle.ack(id="m003")
        msg = "This is a message"
        s = sprinkle.send(msg, destination="gardening")
        t = str(a) + str(s)
        self.d.load(t)
        self.failUnlessEqual(list(self.d),
        [("ACK", {"message-id":"m003"}, "")
        ,("SEND",{"content-length": str(len(msg)) 
        ,"destination":"gardening"},msg)])
 
    def test_003(self):
        """ Simple error trapping"""
        f = sprinkle.connect(login="siegfried",passcode="canine")
        self.d.load("DODGY\nbroken header\n\n\x00")
        self.d.load(str(f))
        self.d.load(str(sprinkle.subscribe("bingo","auto")))
        self.d.load(str(sprinkle.disconnect()))
        self.failUnlessEqual(self.d.next(),
        ("CONNECT",{"login": "siegfried","passcode": "canine"},""))
        self.failUnlessEqual(self.d.next(),
        ("SUBSCRIBE",{"destination": "bingo","ack": "auto"},""))
        self.failUnlessEqual(self.d.next(), ("DISCONNECT",{},"\n"))

class TestMessageManager(unittest.TestCase):
    """ Concerning the MessageManager class"""
    
    def setUp(self):
        self.mm = sprinkle.MessageManager(QUEUE_ROOT)

    def test_001(self):
        """ make sure messages can't go to non-existent queues """
        nukeDir(self.mm.root)
        self.failIf(os.path.exists(self.mm.root))
        self.failIf(self.mm.submit("all"
        , "Hello, I'm a hopeless newb!\n"))

    def test_002(self):
        """ Create a default queue, as a sysadmin would """
        makeDir(os.path.join(self.mm.root, "all"))
        self.failUnless("all" in self.mm.queues())

    def test_003(self):
        """ Place a message. Check it's there """
        makeDir(os.path.join(self.mm.root, "all"))
        id = self.mm.submit("all", "Hello, I'm a hopeless newb!\n")
        self.failUnless(id)
        self.failUnless(("all", id) in
        list([self.mm.msgInfo(r) for r in self.mm.messages()]))

    def tearDown(self):
        nukeDir(self.mm.root)
        pass

class TestUserManager(unittest.TestCase):
    """ Concerning the UserManager class"""
    
    def deliverMessage(self, user, queue, content):
        """ A white-box way of delivering a message to a user """
        dir = os.path.join(self.um.root,user)
        f,p = tempfile.mkstemp(".msg", queue + '-',dir, text = True)
        os.write(f,content)
        os.close(f)
        return p

    def setUp(self):
        self.um = sprinkle.UserManager(USER_ROOT)
        makeDir(self.um.root)
        self.um.configure()

    def test_001(self):
        """ From a blank tree, make sure users can't be registered """
        nukeDir(self.um.root)
        self.failIf(self.um.add("siegfried", "password", "gardening",
        "auto"))

    def test_002(self):
        """ Add a user. Check he can subscribe """
        makeUser("siegfried", "password", ("all",))
        self.um.configure()
        self.failUnless(self.um.add("siegfried", "password", "all", "auto"))
        self.failUnless(self.um.add("siegfried","password","gardening","auto"))

    def test_003(self):
        """ Test reporting of subscribers """
        makeUser("siegfried", "password", ("all", "gardening"))
        makeUser("tristan", "password", ("all","pottery"))
        self.um.configure()
        self.failUnlessEqual(self.um.subscribers("all"), ["tristan", "siegfried"])
        self.failUnlessEqual(self.um.subscribers("gardening"), ["siegfried"])
        self.failUnlessEqual(self.um.subscribers("pottery"), ["tristan"])

    def test_004(self):
        """ Test the remove method"""
        makeUser("siegfried", "password", ("all", "gardening"))
        makeUser("tristan", "password", ("all","pottery"))
        self.um.configure()
        self.failUnless(self.um.remove("siegfried", "password",
        "gardening"))
        self.failIf(self.um.remove("tristan", "notthepassword",
        "pottery"))
        # Make sure changes have stuck
        self.um.configure()
        self.failUnlessEqual(self.um.subscribers("gardening"), [])
        self.failUnlessEqual(self.um.subscribers("pottery"), ["tristan"])
        # Restore Siegfried's subscription
        self.failUnless(self.um.add("siegfried","password","gardening","auto"))
        self.um.configure()
        
    def test_005(self):
        """ Test the pending method - auto acknowledge"""
        self.deliverMessage("siegfried","gardening"
        ,"When do I repot my green beans?")
        self.failUnlessEqual(len(list(self.um.pending("siegfried"))), 1)
        self.failUnlessEqual(len(list(self.um.pending("siegfried"))), 0)

    def test_006(self):
        """ Test the pending method - client acknowledge"""
        makeUser("siegfried", "password", ("all", "gardening"),
        ack="client")
        self.um.configure()
        self.deliverMessage("siegfried","gardening"
        ,"When do I repot my green beans?")
        self.failUnlessEqual(len(list(self.um.pending("siegfried"))), 1)
        self.failUnlessEqual(len(list(self.um.pending("siegfried"))), 1)

    def test_007(self):
        """ Test the acknowledge method - depends on test_006"""
        self.failUnlessEqual(len(list(self.um.pending("siegfried"))), 1)
        path = self.deliverMessage("siegfried","gardening"
        ,"When do I repot my runner beans?")
        msgId = os.path.basename(path)
        self.failUnlessEqual(len(list(self.um.pending("siegfried"))), 2)
        self.failUnless(self.um.acknowledge("siegfried", msgId))
        self.failUnlessEqual(len(list(self.um.pending("siegfried"))), 1)
        self.failIf(self.um.acknowledge("siegfried", msgId))

    def test_999(self):
        nukeDir(self.um.root)

    def tearDown(self):
        #nukeDir(self.um.root)
        pass

class TestWorker(unittest.TestCase):
    """ Concerning the Worker class"""
    
    sfn = "./testsocket"

    def setUp(self):
        try: os.remove(self.sfn)
        except OSError: pass
        self.w = sprinkle.Worker(os.getpid())
        self.w.opts = optparse.Values()
        self.w.opts.debug = False
        self.w.um = sprinkle.UserManager(home = USER_ROOT)
        makeUser("siegfried", "canine", ["all","gardening"])
        makeUser("tristan", "french", ["all","cookery"])
        self.w.um.configure()
        self.w.s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.w.s.bind(self.sfn)
        self.w.s.listen(1)
        
    def test_001(self):
        """ test the connect transaction """
        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(self.w.s.getsockname())
        client.sendall(str(sprinkle.connect(login="siegfried",passcode="canine")))
        client.sendall(str(sprinkle.disconnect()))

        self.w()
        reply = client.recv(1024)
        self.failUnlessEqual(reply, str(sprinkle.connected(os.getpid()))
         + str(sprinkle.error("DISCONNECT","disconnection notice")))

    def test_002(self):
        """ test the subscribe transaction
            result: a call to UserManager.add"""
        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(self.w.s.getsockname())
        client.sendall(str(sprinkle.connect(login="siegfried",passcode="canine")))
        client.sendall(str(sprinkle.subscribe("bingo","auto")))
        client.sendall(str(sprinkle.disconnect()))

        self.w()
        reply = client.recv(1024)
        self.failUnlessEqual(reply, str(sprinkle.connected(os.getpid()))
         + str(sprinkle.error("DISCONNECT","disconnection notice")))
        self.failUnless("siegfried" in self.w.um.subscribers("bingo"))

    def test_003(self):
        """ test the unsubscribe transaction
            result: a call to UserManager.remove"""
        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(self.w.s.getsockname())
        client.sendall(str(sprinkle.connect(login="siegfried",passcode="canine")))
        client.sendall(str(sprinkle.subscribe("bingo","auto")))
        client.sendall(str(sprinkle.unsubscribe("bingo")))
        client.sendall(str(sprinkle.disconnect()))

        self.w()
        reply = client.recv(1024)
        self.failUnlessEqual(reply, str(sprinkle.connected(os.getpid()))
         + str(sprinkle.error("DISCONNECT","disconnection notice")))
        self.failIf("siegfried" in self.w.um.subscribers("bingo"))

    def test_004(self):
        """ test the subscribe transaction with a receipt request
            result: a call to UserManager.add and a receipt"""
        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(self.w.s.getsockname())
        client.sendall(str(sprinkle.connect(login="siegfried",passcode="canine")))
        client.sendall(str(
        sprinkle.subscribe("bingo","auto").requestReceipt("r01")))
        client.sendall(str(sprinkle.disconnect()))

        self.w()
        reply = client.recv(1024).split('\x00')
        self.failUnless(reply[0].startswith("CONNECTED"))
        self.failUnless(reply[1].startswith("RECEIPT\nreceipt-id: r01\n"))
        self.failUnless("siegfried" in self.w.um.subscribers("bingo"))

    def test_005(self):
        """ test a transaction which results in a commit"""
        class dummy(object):
            submitted = 0        
            def submit(self,*args): dummy.submitted += 1
        self.w.mm = dummy()

        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(self.w.s.getsockname())
        client.sendall(str(sprinkle.connect(login="siegfried",passcode="canine")))
        client.sendall(str(sprinkle.begin("t01")))
        client.sendall(str(
        sprinkle.send("hi!", "bingo").joinTransaction("t01")))
        client.sendall(str(
        sprinkle.send("hi!", "bingo").joinTransaction("t01")))
        client.sendall(str(
        sprinkle.send("hi!", "bingo").joinTransaction("t01")))
        client.sendall(str(sprinkle.commit("t01")))
        client.sendall(str(sprinkle.disconnect()))
        self.w.mm = dummy()
        self.w()
        reply = client.recv(1024)
        self.failUnlessEqual(self.w.mm.submitted, 3)

    def test_006(self):
        """ test a transaction which results in an abort"""
        class dummy(object):
            submitted = 0
            def submit(self, *args):dummy.submitted += 1
        self.w.mm = dummy()

        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(self.w.s.getsockname())
        client.sendall(str(sprinkle.connect(login="siegfried",passcode="canine")))
        client.sendall(str(sprinkle.begin("t01")))
        client.sendall(str(
        sprinkle.send("hi!", "bingo").joinTransaction("t01")))
        client.sendall(str(
        sprinkle.send("hi!", "bingo").joinTransaction("t01")))
        client.sendall(str(
        sprinkle.send("hi!", "bingo").joinTransaction("t01")))
        client.sendall(str(sprinkle.abort("t01")))
        client.sendall(str(sprinkle.disconnect()))

        self.w()
        reply = client.recv(1024)
        self.failUnlessEqual(self.w.mm.submitted, 0)

    def test_007(self):
        """ test that users can't be registered to non-existent queues """
        #TODO!
        #makeUser("siegfried", "password", ("all",))
        #self.um.configure()
        #self.failIf(self.um.add("siegfried", "password", "gardening",
        #"auto"))
        pass

    def tearDown(self):
        self.w.s.shutdown(2)
        self.w.s.close()
        try: os.remove(self.sfn)
        except OSError: pass
        nukeDir(self.w.um.root)

class TelnetTester(unittest.TestCase):

    def setUp(self):
        self.help = TestHelper()
        self.t = telnetlib.Telnet("localhost", 12000)

    def test_001(self):
        """ Test skipping connection """
        self.help.makeDir(UserManager.root)
        self.help.makeUser("siegfried", "password", ("all",))
        self.t.write("SUBSCRIBE\ndestination: archery\n"
        "ack: client\n\n\x00")
        reply = self.t.read_until("\x00")
        self.failUnless(reply.startswith("ERROR\n"))

    def test_002(self):
        """ Test invalid login """
        self.t.write("CONNECT\nlogin: notauser\n"
        "passcode: notapassword\n\n\x00")
        reply = self.t.read_until("\x00")
        self.failUnless(reply.startswith("ERROR\n"))

    def test_003(self):
        """ Test invalid password """
        self.t.write("CONNECT\nlogin: siegfried\n"
        "passcode: notsiegfriedspassword\n\n\x00")
        reply = self.t.read_until("\x00")
        self.failUnless(reply.startswith("ERROR\n"))

    def test_004(self):
        """ Test a valid connection """
        self.t.write("CONNECT\nlogin: siegfried\n"
        "passcode: password\n\n\x00")
        reply = self.t.read_until("\x00")
        self.t.write("SUBSCRIBE\ndestination: gardening\n"
        "receipt:test_001\n\n\x00")
        reply = self.t.read_until('\x00')
        self.failUnless(reply.startswith(
        "RECEIPT\nreceipt-id: test_001\n\n"))

    def tost_005(self):
        self.t.write("CONNECT\nlogin: siegfried\n"
        "passcode: password\n\n\x00")
        self.t.write("SUBSCRIBE\ndestination: gardening\n"
        "\n\x00")

    def tearDown(self):
        self.t.write("DISCONNECT\n\n\x00")
        time.sleep(0.5)
        self.t.close()

def main():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestFormat))
    suite.addTest(unittest.makeSuite(TestDeframer))
    suite.addTest(unittest.makeSuite(TestMessageManager))
    suite.addTest(unittest.makeSuite(TestUserManager))
    suite.addTest(unittest.makeSuite(TestWorker))
    unittest.TextTestRunner(verbosity=1).run(suite)
    return 0

if __name__ == "__main__":   
    sys.exit(main())
